#!/usr/bin/python
# -*- coding: utf-8 -*-

#  Copyright © 2013-2014  B. Clausius <barcc@gmx.de>
#
#  This program is free software: you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation, either version 3 of the License, or
#  (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#
#  You should have received a copy of the GNU General Public License
#  along with this program.  If not, see <http://www.gnu.org/licenses/>.

from __future__ import print_function, division

import sys, os
import random

from gi.repository import GLib
from gi.repository import Gio
from gi.repository import Gtk
from gi.repository import Gedit

data_dir = os.path.join(os.path.dirname(__file__), 'data')

class _Actions (object):
    pass
    
def print_diff(lista, listb):
    import difflib
    for line in difflib.unified_diff(lista, listb, lineterm=''):
        print(line)
        
def get_actions():
    window = Gedit.App.get_default().get_active_window()
    uimanager = window.get_ui_manager()
    actiongroups = uimanager.get_action_groups()
    actions = _Actions()
    for actiongroup in actiongroups:
        if actiongroup.get_name().startswith('ProjectsPlugin'):
            for action in actiongroup.list_actions():
                name = action.get_name()
                setattr(actions, name, action)
    return actions
                
def _get_widget( widget, name):
    widgetlist = [widget]
    while widgetlist:
        widget = widgetlist.pop(0)
        _name = Gtk.Buildable.get_name(widget)
        if _name == name:
            return widget
        elif isinstance(widget, Gtk.Container):
            widgetlist += widget.get_children()
    return None
    
def get_open_filenames():
    window = Gedit.App.get_default().get_active_window()
    filenames = []
    for doc in window.get_documents():
        path = doc.get_location().get_path()
        filenames.append(os.path.relpath(path, data_dir))
    return filenames
    
def get_open_projects():
    window = Gedit.App.get_default().get_active_window()
    panel = window.get_side_panel()
    widget = _get_widget(panel, 'treeview_open_projects')
    model = widget and widget.get_model()
    return model
    
def get_all_projects():
    window = Gedit.App.get_default().get_active_window()
    panel = window.get_side_panel()
    widget = _get_widget(panel, 'treeview_projects')
    model = widget and widget.get_model()
    return model
    
def get_active_path():
    doc = Gedit.App.get_default().get_active_window().get_active_document()
    location = doc and doc.get_location()
    return location and location.get_path()
    
def iter_model(model, titer=None):
    titer = model.iter_children(titer)
    while titer:
        yield titer
        for citer in iter_model(model, titer):
            yield citer
        titer = model.iter_next(titer)
        
def get_model_paths(model, titer=None):
    paths = []
    titer = model.iter_children(titer)
    for titer in iter_model(model):
        path = model[titer][1]
        assert path.startswith('file://')
        path = path[7:]
        paths.append(os.path.relpath(path, data_dir))
    return paths
    
def get_open_project_paths():
    model = get_open_projects()
    return get_model_paths(model)
    
def get_all_project_paths():
    model = get_all_projects()
    return get_model_paths(model)
    
def select_first_project():
    window = Gedit.App.get_default().get_active_window()
    panel = window.get_side_panel()
    widget = _get_widget(panel, 'treeview_projects')
    widget.set_cursor(Gtk.TreePath(), None, False)
    
def select_project(project):
    window = Gedit.App.get_default().get_active_window()
    panel = window.get_side_panel()
    widget = _get_widget(panel, 'treeview_projects')
    model = widget.get_model()
    projectpath = get_uri(project)
    for titer in iter_model(model):
        if projectpath == model[titer][1]:
            tpath = model.get_path(titer)
            widget.expand_to_path(tpath)
            widget.set_cursor(tpath, None, False)
            return True
    return False
    
def get_path(*basenames):
    return os.path.join(data_dir, *basenames)
    
def get_uri(*basenames):
    return 'file://' + os.abspath(get_path(*basenames))
    
def get_random_paths(dirname, breadth, maxdepth):
    if maxdepth <= 0:
        return []
    directories = []
    for i in range(random.randrange(breadth)):
        dirname_ = os.path.join(dirname, 'd%s' % (i+1))
        directories.append(dirname_)
        directories += get_random_paths(dirname_, breadth, maxdepth-1)
    return directories
    
def get_location(basename):
    return Gio.File.new_for_path(get_path(basename))
    
def new_doc():
    window = Gedit.App.get_default().get_active_window()
    window.create_tab(True)
    
def open_doc(basename):
    location = get_location(basename)
    window = Gedit.App.get_default().get_active_window()
    window.create_tab_from_location(location, None, 0, 0, False, True)
    
def close_all():
    window = Gedit.App.get_default().get_active_window()
    window.close_all_tabs()
    assert len(window.get_views()) == 0
    
def close_active():
    window = Gedit.App.get_default().get_active_window()
    tab = window.get_active_tab()
    window.close_tab(tab)
    
def open_project(project):
    if select_project(project):
        get_actions().action_open_project.activate()
        
def close_project(project):
    if select_project(project):
        get_actions().action_close_project.activate()
        
def open_file(project):
    action_open_file = get_actions().action_open_file
    action_open_file.test_directory = get_uri(project)
    if select_project(os.path.dirname(project)):
        action_open_file.activate()
        
def add_project(project):
    action_add_directory = get_actions().action_add_directory
    action_add_directory.test_directory = get_uri(project)
    action_add_directory.activate()
    
def add_parent_project(project):
    if select_project(project):
        get_actions().action_add_parent.activate()
        
def remove_project(project):
    if select_project(project):
        get_actions().action_remove.activate()
        
def remove_projects():
    get_actions().action_remove_all.activate()
    assert len(get_open_projects()) == 0
    assert len(get_all_projects()) == 0
    
def remove_files():
    for p, dirs, files in os.walk(get_path('random'), False):
        for f in files:
            os.remove(os.path.join(p, f))
        os.rmdir(p)
        
def clear():
    close_all()
    remove_projects()
    remove_files()
    
def ls(module):
    print(module.__name__)
    i = 0
    for name in sorted(dir(module)):
        if name.startswith('test_'):
            i += 1
            print(' {:3}. {}'.format(i, name))
            
_run_queue = []
_success = []

def _run_idle():
    name, func, args = _run_queue.pop(0)
    if name is not None:
        print('**** Running test: %s' % name)
        _success[0] += 1
    try:
        funcs = func(*args)
    except Exception:
        _success[1] += 1
        sys.excepthook(*sys.exc_info())
    else:
        if funcs is not None:
            if not isinstance(funcs, list):
                funcs = [funcs]
            for func in reversed(funcs):
                func, args = func if isinstance(func, tuple) else (func, [])
                _run_queue.insert(0, (None, func, args))
    return bool(_run_queue)
    
def run(module, testname):
    if not _run_queue:
        GLib.idle_add(_run_idle, priority=GLib.PRIORITY_LOW)
    _success[:] = [0, 0]
    _run_queue.append((None, print, ['#### Running suite: %s' % os.path.basename(module.__file__)]))
    i = 0
    for name in sorted(dir(module)):
        if name.startswith('test_'):
            i += 1
            if testname in [None, name, i]:
                print(' {:3}. {}'.format(i, name))
                _run_queue.append((None, clear, []))
                _run_queue.append((name, getattr(module, name), []))
    def _result():
        print('#### {} tests, {} failed\n'.format(*_success))
    _run_queue.append((None, _result, []))
    
    
